/*
 *  bebox.cpp - BeBox specific stuff
 *
 *  Basilisk (C) 1996 Christian Bauer
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define __STDC__ 1
#define __GNU_LIBRARY__
#include <getopt.h>
#undef __GNU_LIBRARY__
#undef __STDC__

#ifndef __BEOS__
#error Compiling bebox.cpp, but __BEOS__ unset.
#endif

extern "C" {
#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"
#include "memory.h"
#include "newcpu.h"
#include "patches.h"
#include "sony.h"
#include "filedisk.h"
#include "via.h"
#include "debug.h"
#include "gui.h"
#include "xwin.h"
#include "zfile.h"
}


// Mac Classic screen dimensions
const int DISPLAY_X = 512;
const int DISPLAY_Y = 342;
const BRect DisplayFrame = BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1);

// Messages
const ulong MSG_REDRAW = 'draw';
const ulong MSG_FLOPPY = 'flop';
const ulong MSG_DEBUG = 'dbug';

// Pointer to Mac screen buffer
UBYTE *MacScreenBuf;

// Tables for converting 1 Bit -> 8 Bit bitmaps
ULONG ExpTabHi[0x100];
ULONG ExpTabLo[0x100];

// Be cursor image
UBYTE the_cursor[68] = {16, 1};

bool mouse_in_view = TRUE;	// Flag: Mouse pointer within bitmap view

int quit_program;


class MacWindow;
class BitmapView;
class Emulator;


/*
 *  The BeOS application object
 */

class Basilisk : public BApplication {
public:
	Basilisk();
	virtual void ArgvReceived(int argc, char **argv);
	virtual void ReadyToRun(void);
	virtual bool QuitRequested(void);
	virtual void AboutRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	MacWindow *main_window;
	Emulator *the_emulator;
};


/*
 *  The window in which the Mac graphics are displayed, handles I/O
 */

class MacWindow : public BWindow {
public:
	MacWindow(BRect frame);
	virtual bool QuitRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	BBitmap *the_bitmap;
	BitmapView *main_view;
};


/*
 *  A simple view class for blitting a bitmap on the screen
 */

class BitmapView : public BView {
public:
	BitmapView(BRect frame, BBitmap *bitmap);
	virtual void Draw(BRect update);
	virtual	void MouseMoved(BPoint point, ulong transit, BMessage *message);

private:
	BBitmap *the_bitmap;
};


/*
 *  For running the emulation in its own thread
 */

class Emulator {
public:
	Emulator() {thread_running = FALSE;}
	void Run(void);
	void Quit(void);

private:
	static long thread_invoc(void *obj);
	void thread_func(void);

	thread_id the_thread;
	bool thread_running;
};


/*
 *  Create application object and start it
 */

main()
{
	Basilisk *the_app;
	
	the_app = new Basilisk();
	the_app->Run();
	delete the_app;
	
	return 0;
}


/*
 *  Application constructor: Initialize member variables
 */

Basilisk::Basilisk() : BApplication('McEm')
{
	the_emulator = NULL;
	main_window = NULL;
}


/*
 *  Parse command line options
 */

void usage(void)
{
	printf("Basilisk\n");
	printf("Summary of command-line options:\n");
	printf("  -h                 : Print help\n");
	printf("  -f n               : Set the frame rate to n milliseconds\n");
	printf("  -D                 : Start up the built-in debugger\n");
	printf("  -E                 : Boot from ROM .EDisk\n");
	printf("  -[0123] filename   : Use file instead of fd[0123].image as filedisks\n");
	printf("  -r filename        : Use file as ROM image instead of 'ClassicROM'\n");
    printf("  -p filename        : Use filename to save PRAM contents\n");
}

void Basilisk::ArgvReceived(int argc, char **argv)
{
	int c;

	while(((c = getopt(argc, argv, "DEf:h0:1:2:3:r:p:")) != EOF)) switch(c) {
     case 'h': usage();	exit(0);

     case '0': strncpy(df0, optarg, 255); df0[255] = 0;	break;
     case '1': strncpy(df1, optarg, 255); df1[255] = 0; break;
     case '2': strncpy(df2, optarg, 255); df2[255] = 0; break;
     case '3': strncpy(df3, optarg, 255); df3[255] = 0; break;
     case 'r': strncpy(romfile, optarg, 255); romfile[255] = 0; break;
     case 'p': strncpy(pramfile, optarg, 255); pramfile[255] = 0; break;		    
	
     case 'f': framerate = atoi(optarg); break;
     case 'D': use_debugger = 1; break;
     case 'E': boot_edisk = 1; break;
	}
}


/*
 *  Arguments processed, create and start emulation
 */

void Basilisk::ReadyToRun(void)
{
	// Initialize everything	
	memory_init();
	load_pram();
	sony_init();
	filedisk_init();
	init_m68k();
	m68k_reset();
	if (use_debugger && debuggable())
		activate_debugger();

	// Get address of Mac screen buffer
	MacScreenBuf = get_real_address(0x3fa700);

	// Initialize ExpTab
	for (int i=0; i<0x100; i++) {
		ExpTabHi[i] = ((i & 0x80) ? 0 : 0xff000000)
					| ((i & 0x40) ? 0 : 0x00ff0000)
					| ((i & 0x20) ? 0 : 0x0000ff00)
					| ((i & 0x10) ? 0 : 0x000000ff);
		ExpTabLo[i] = ((i & 0x08) ? 0 : 0xff000000)
					| ((i & 0x04) ? 0 : 0x00ff0000)
					| ((i & 0x02) ? 0 : 0x0000ff00)
					| ((i & 0x01) ? 0 : 0x000000ff);
	}

	// Create application main menu
	BPopUpMenu *menu = new BPopUpMenu("", FALSE, FALSE);
	menu->AddItem(new BMenuItem("About Basilisk...", new BMessage(B_ABOUT_REQUESTED)));
	menu->AddItem(new BMenuItem("Mount Floppy Disk", new BMessage(MSG_FLOPPY)));
	menu->AddItem(new BMenuItem("Debugger", new BMessage(MSG_DEBUG)));
	menu->AddItem(new BMenuItem("Quit Basilisk", new BMessage(B_QUIT_REQUESTED), 'Q'));
	menu->SetTargetForItems(be_app);
	SetMainMenu(menu);

	// Open window
	main_window = new MacWindow(DisplayFrame);

	// Start emulation
	the_emulator = new Emulator;
	the_emulator->Run();
}


/*
 *  Quit requested (either by menu or by closing the window)
 */

bool Basilisk::QuitRequested(void)
{
	if (BApplication::QuitRequested()) {

		// Stop emulation
		if (the_emulator) {
			the_emulator->Quit();
			delete the_emulator;
		}

		dump_counts();
		filedisk_exit();
		sony_exit();
		save_pram();
		zfile_exit();
		return TRUE;
	}
	return FALSE;
}


/*
 *  Display "about" window
 */

void Basilisk::AboutRequested(void)
{
	char str[256];

	sprintf(str, "Basilisk V0.4 by Christian Bauer\n<cbauer@iphcip1.physik.uni-mainz.de>");
	BAlert *the_alert = new BAlert("", str, "OK");
	the_alert->Go();
}


/*
 *  Message received
 */

void Basilisk::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_FLOPPY:
			mount_floppy = 1;
			break;

		case MSG_DEBUG:
			activate_debugger();
			break;

		default:
			BApplication::MessageReceived(msg);
			break;
	}
}


/*
 *  MacWindow constructor
 */

MacWindow::MacWindow(BRect frame) : BWindow(frame, "Basilisk", B_TITLED_WINDOW, B_NOT_RESIZABLE)
{
	// Move window to right position
	Lock();
	MoveTo(80, 60);

	// Allocate bitmap
	the_bitmap = new BBitmap(DisplayFrame, B_COLOR_8_BIT);

	// Create bitmap view
	main_view = new BitmapView(DisplayFrame, the_bitmap);
	AddChild(main_view);
	main_view->MakeFocus();

	// Show window
	Unlock();
	Show();

	// Start redrawing loop
	PostMessage(MSG_REDRAW);
}


/*
 *  Closing the window quits the program
 */

bool MacWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}


/*
 *  Handles redraw messages
 */

void MacWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_REDRAW:

			double start_time = system_time();

			// Redraw bitmap unless ScrollLock is down
			if (!(modifiers() & B_SCROLL_LOCK)) {

				// Convert 1 Bit -> 8 Bit
				UBYTE *p = MacScreenBuf;
				ULONG *q = (ULONG *)the_bitmap->Bits();
				for (int i=0; i<DISPLAY_X*DISPLAY_Y/8; i++) {
					*q++ = ExpTabHi[*p];
					*q++ = ExpTabLo[*p++];
				}

				Lock();
				main_view->Draw(DisplayFrame);
				Unlock();
			}

			// Has the Mac started?
			if (get_long(0xcfc) == 'WLSC') {
				if (mouse_in_view) {

					// Poll mouse buttons and keyboard
					BPoint cursor;
					ulong buttons;
					key_info keys;
					Lock();
					main_view->GetMouse(&cursor, &buttons, FALSE);
					main_view->GetKeys(&keys, TRUE);
					Unlock();
					mousebutton = buttons & B_PRIMARY_MOUSE_BUTTON;
					memcpy(key_states, keys.key_states, sizeof(key_states));

					// Set cursor image
					if (memcmp(the_cursor+4, get_real_address(0x844), 64)) {
						memcpy(the_cursor+4, get_real_address(0x844), 64);	// Cursor image
						the_cursor[2] = get_byte(0x885);	// Hotspot
						the_cursor[3] = get_byte(0x887);
						be_app->SetCursor(the_cursor);
					}
				}

				// Trigger Mac VBL interrupt
				TriggerVBL();
			}

			// Wait to maintain constant frame rate
			double wait_time = framerate * 1000 - (system_time() - start_time);
			if (wait_time > 0)
				snooze(wait_time);

			// Continue redrawing loop
			PostMessage(MSG_REDRAW);
			break;

		default:
			BWindow::MessageReceived(msg);
	}
}


/*
 *  Bitmap view constructor
 */

BitmapView::BitmapView(BRect frame, BBitmap *bitmap) : BView(frame, "", B_FOLLOW_NONE, B_WILL_DRAW)
{
	the_bitmap = bitmap;
}


/*
 *  Blit the bitmap
 */

void BitmapView::Draw(BRect update)
{
	DrawBitmap(the_bitmap, update, update);
}


/*
 *  Mouse moved
 */

void BitmapView::MouseMoved(BPoint point, ulong transit, BMessage *message)
{
	switch (transit) {
		case B_ENTERED_VIEW:
			mouse_in_view = TRUE;
			be_app->SetCursor(the_cursor);
			mousex = point.x;
			mousey = point.y;
			break;
		case B_EXITED_VIEW:
			mouse_in_view = FALSE;
			be_app->SetCursor(B_HAND_CURSOR);
			break;
		case B_INSIDE_VIEW:
			mousex = point.x;
			mousey = point.y;
			break;
	}
}


/*
 *  Start main emulation thread
 */

void Emulator::Run(void)
{
	// Start the emulation thread
	the_thread = spawn_thread(thread_invoc, "Mac 68000", B_NORMAL_PRIORITY, this);
	resume_thread(the_thread);
	thread_running = TRUE;
}


/*
 *  Stop main emulation thread
 */

void Emulator::Quit(void)
{
	// Kill the thread if it is running
	if (thread_running) {
		kill_thread(the_thread);
		thread_running = FALSE;
	}
}


/*
 *  The thread's main function
 */

long Emulator::thread_invoc(void *obj)
{
	((Emulator *)obj)->thread_func();
	be_app->PostMessage(B_QUIT_REQUESTED);
	return 0;
}

void Emulator::thread_func(void)
{
	m68k_go(1);
	thread_running = FALSE;
}


/*
 *  Mac floppy ejected, inform user
 */

void gui_disk_unmounted(int num)
{
	BAlert *the_alert = new BAlert("", "The floppy disk was 'ejected'.", "OK");
	the_alert->Go();
}


int debuggable(void)
{
    return 1;
}
