/*
 *  memory.c - Memory management
 *
 *  Basilisk (C) 1996 Christian Bauer
 */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "memory.h"
#include "patches.h"
#include "newcpu.h"
#include "zfile.h"

int buserr;
addrbank membanks[65536];

/* Default memory access functions */

int default_check(CPTR a, ULONG b)
{
    return 0;
}

UBYTE *default_xlate(CPTR a)
{
    fprintf(stderr, "Your Mac program just did something terribly stupid\n");
    return 0;
}

UWORD default_awget(CPTR addr)
{
    default_xlate(addr);
    return 0;
}

ULONG default_alget(CPTR addr)
{
    default_xlate(addr);
    return 0;
}

/* A dummy bank that only contains zeros */

static ULONG dummy_lget(CPTR) REGPARAM;
static UWORD dummy_wget(CPTR) REGPARAM;
static UBYTE dummy_bget(CPTR) REGPARAM;
static void  dummy_lput(CPTR, ULONG) REGPARAM;
static void  dummy_wput(CPTR, UWORD) REGPARAM;
static void  dummy_bput(CPTR, UBYTE) REGPARAM;
static int   dummy_check(CPTR addr, ULONG size) REGPARAM;

ULONG dummy_lget(CPTR addr)
{
    if (illegal_mem) printf("Illegal lget at %08lx pc=%08lx\n",addr,m68k_getpc());
    return 0;
}

UWORD dummy_wget(CPTR addr)
{
    if (illegal_mem) printf("Illegal wget at %08lx pc=%08lx\n",addr,m68k_getpc());
    return 0;
}

UBYTE dummy_bget(CPTR addr)
{
    if (illegal_mem) printf("Illegal bget at %08lx pc=%08lx\n",addr,m68k_getpc());
    return 0;
}

void dummy_lput(CPTR addr, ULONG l)
{
    if (illegal_mem) printf("Illegal lput at %08lx pc=%08lx\n",addr,m68k_getpc());
}
void dummy_wput(CPTR addr, UWORD w)
{
    if (illegal_mem) printf("Illegal wput at %08lx pc=%08lx\n",addr,m68k_getpc());
}
void dummy_bput(CPTR addr, UBYTE b)
{
    if (illegal_mem) printf("Illegal bput at %08lx pc=%08lx\n",addr,m68k_getpc());
}

int dummy_check(CPTR addr, ULONG size)
{
    if (illegal_mem) printf("Illegal check at %08lx pc=%08lx\n",addr,m68k_getpc());
    return 0;
}

/* Main memory */

static UBYTE *mainmemory;

static ULONG mainmem_alget(CPTR) REGPARAM;
static UWORD mainmem_awget(CPTR) REGPARAM;

static ULONG mainmem_lget(CPTR) REGPARAM;
static UWORD mainmem_wget(CPTR) REGPARAM;
static UBYTE mainmem_bget(CPTR) REGPARAM;
static void  mainmem_lput(CPTR, ULONG) REGPARAM;
static void  mainmem_wput(CPTR, UWORD) REGPARAM;
static void  mainmem_bput(CPTR, UBYTE) REGPARAM;

static int   mainmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE *mainmem_xlate(CPTR addr) REGPARAM;

ULONG mainmem_alget(CPTR addr)
{
#ifdef __BEOS__
	return *(ULONG *)(mainmemory + (addr & 0xffffff));
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
#endif
}

UWORD mainmem_awget(CPTR addr)
{
#ifdef __BEOS__
	return *(UWORD *)(mainmemory + (addr & 0xffffff));
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
#endif
}

ULONG mainmem_lget(CPTR addr)
{
#ifdef __BEOS__
	return *(ULONG *)(mainmemory + (addr & 0xffffff));
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
#endif
}

UWORD mainmem_wget(CPTR addr)
{
#ifdef __BEOS__
	return *(UWORD *)(mainmemory + (addr & 0xffffff));
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
#endif
}

UBYTE mainmem_bget(CPTR addr)
{
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    return mainmemory[addr];
}

void mainmem_lput(CPTR addr, ULONG l)
{
#ifdef __BEOS__
	*(ULONG *)(mainmemory + (addr & 0xffffff)) = l;
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    m[0] = l >> 24;
    m[1] = l >> 16;
    m[2] = l >> 8;
    m[3] = l;
#endif
}

void mainmem_wput(CPTR addr, UWORD w)
{
#ifdef __BEOS__
	*(UWORD *)(mainmemory + (addr & 0xffffff)) = w;
#else
    UBYTE *m;
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    m = mainmemory + addr;
    m[0] = w >> 8;
    m[1] = w;
#endif
}

void mainmem_bput(CPTR addr, UBYTE b)
{
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    mainmemory[addr] = b;
}

int mainmem_check(CPTR addr, ULONG size)
{
    addr -= mainmem_start & (mainmem_size-1);
    addr &= mainmem_size-1;
    return (addr + size) < mainmem_size;
}

UBYTE *mainmem_xlate(CPTR addr)
{
    addr -= mainmem_start & mainmem_mask;
    addr &= mainmem_size-1;
    return mainmemory + addr;
}

/* ROM memory */

static UBYTE *rommemory;

static ULONG rommem_alget(CPTR) REGPARAM;
static UWORD rommem_awget(CPTR) REGPARAM;
static ULONG rommem_lget(CPTR) REGPARAM;
static UWORD rommem_wget(CPTR) REGPARAM;
static UBYTE rommem_bget(CPTR) REGPARAM;
static void  rommem_lput(CPTR, ULONG) REGPARAM;
static void  rommem_wput(CPTR, UWORD) REGPARAM;
static void  rommem_bput(CPTR, UBYTE) REGPARAM;
static int   rommem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE *rommem_xlate(CPTR addr) REGPARAM;

ULONG rommem_alget(CPTR addr)
{
#ifdef __BEOS__
	return *(ULONG *)(rommemory + (addr & 0xffffff) - rommem_start);
#else
    UBYTE *m;
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    m = rommemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
#endif
}

UWORD rommem_awget(CPTR addr)
{
#ifdef __BEOS__
	return *(UWORD *)(rommemory + (addr & 0xffffff) - rommem_start);
#else
    UBYTE *m;
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    m = rommemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
#endif
}

ULONG rommem_lget(CPTR addr)
{
#ifdef __BEOS__
	return *(ULONG *)(rommemory + (addr & 0xffffff) - rommem_start);
#else
    UBYTE *m;
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    m = rommemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
#endif
}

UWORD rommem_wget(CPTR addr)
{
#ifdef __BEOS__
	return *(UWORD *)(rommemory + (addr & 0xffffff) - rommem_start);
#else
    UBYTE *m;
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    m = rommemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
#endif
}

UBYTE rommem_bget(CPTR addr)
{
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    return rommemory[addr];
}

void rommem_lput(CPTR addr, ULONG l)
{
    if (illegal_mem) printf("Illegal rommem lput at %08lx pc=%08lx\n",addr,m68k_getpc());
}

void rommem_wput(CPTR addr, UWORD w)
{
    if (illegal_mem) printf("Illegal rommem wput at %08lx pc=%08lx\n",addr,m68k_getpc());
}

void rommem_bput(CPTR addr, UBYTE b)
{
    if (illegal_mem) printf("Illegal rommem lput at %08lx pc=%08lx\n",addr,m68k_getpc());
}

int rommem_check(CPTR addr, ULONG size)
{
    addr -= rommem_start & (rommem_size-1);
    addr &= rommem_size-1;
    return (addr + size) < rommem_size;
}

UBYTE *rommem_xlate(CPTR addr)
{
    addr -= rommem_start & rommem_mask;
    addr &= rommem_size-1;
    return rommemory + addr;
}

/* Emulator private memory */

static UBYTE *emulmemory;

static ULONG emulmem_alget(CPTR) REGPARAM;
static UWORD emulmem_awget(CPTR) REGPARAM;

static ULONG emulmem_lget(CPTR) REGPARAM;
static UWORD emulmem_wget(CPTR) REGPARAM;
static UBYTE emulmem_bget(CPTR) REGPARAM;
static void  emulmem_lput(CPTR, ULONG) REGPARAM;
static void  emulmem_wput(CPTR, UWORD) REGPARAM;
static void  emulmem_bput(CPTR, UBYTE) REGPARAM;

static int   emulmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE *emulmem_xlate(CPTR addr) REGPARAM;

ULONG emulmem_alget(CPTR addr)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
}

UWORD emulmem_awget(CPTR addr)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
}

ULONG emulmem_lget(CPTR addr)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    return (((ULONG)m[0] << 24) + ((ULONG)m[1] << 16) 
	    + ((ULONG)m[2] << 8) + ((ULONG)m[3]));
}

UWORD emulmem_wget(CPTR addr)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    return ((UWORD)m[0] << 8) + m[1];
}

UBYTE emulmem_bget(CPTR addr)
{
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    return emulmemory[addr];
}

void emulmem_lput(CPTR addr, ULONG l)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    m[0] = l >> 24;
    m[1] = l >> 16;
    m[2] = l >> 8;
    m[3] = l;
}

void emulmem_wput(CPTR addr, UWORD w)
{
    UBYTE *m;
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    m = emulmemory + addr;
    m[0] = w >> 8;
    m[1] = w;
}

void emulmem_bput(CPTR addr, UBYTE b)
{
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    emulmemory[addr] = b;
}

int emulmem_check(CPTR addr, ULONG size)
{
    addr -= emulmem_start & (emulmem_size-1);
    addr &= emulmem_size-1;
    return (addr + size) < emulmem_size;
}

UBYTE *emulmem_xlate(CPTR addr)
{
    addr -= emulmem_start & emulmem_mask;
    addr &= emulmem_size-1;
    return emulmemory + addr;
}


static int load_rom(void)
{
	int i;
	FILE *f = zfile_open(romfile, "rb");

    if (f == NULL) {	
    	fprintf(stderr, "No Macintosh ROM found.\n");
	return 0;
    }

    i = fread(rommemory, 1, rommem_size, f);
	if (i != rommem_size) {
		fprintf(stderr, "Error while reading ROM.\n");
		return 0;
    }
    zfile_close(f);
    return 1;
}

/* Address banks */

addrbank dummy_bank = {
    default_alget, default_awget,
    dummy_lget, dummy_wget, dummy_bget,
    dummy_lput, dummy_wput, dummy_bput,
    default_xlate, dummy_check
};

addrbank mainmem_bank = {
    mainmem_alget, mainmem_awget,
    mainmem_lget, mainmem_wget, mainmem_bget,
    mainmem_lput, mainmem_wput, mainmem_bput,
    mainmem_xlate, mainmem_check
};

addrbank rommem_bank = {
    rommem_alget, rommem_awget,
    rommem_lget, rommem_wget, rommem_bget,
    rommem_lput, rommem_wput, rommem_bput,
    rommem_xlate, rommem_check
};

addrbank emulmem_bank = {
    emulmem_alget, emulmem_awget,
    emulmem_lget, emulmem_wget, emulmem_bget,
    emulmem_lput, emulmem_wput, emulmem_bput,
    emulmem_xlate, emulmem_check
};

void memory_init(void)
{
    int i;
    buserr = 0;

	mainmemory = malloc(mainmem_size);
	rommemory = malloc(rommem_size);
	emulmemory = malloc(emulmem_size);

    memset(mainmemory, 0, mainmem_size);
    memset(emulmemory, 0, emulmem_size);

	load_rom();
	patch_rom(rommemory);

    for(i = 0; i < 65536; i++)
	membanks[i] = dummy_bank;

	map_banks(&mainmem_bank, 0x00, 0x40);
	map_banks(&rommem_bank, 0x40, 0x08);
	map_banks(&emulmem_bank, 0x60, 0x01);
	map_banks(&via_bank, 0xef, 0x01);
}

void map_banks(addrbank *bank, int start, int size)
{
    int bnr;
    int hioffs = 0;
#if CPU_LEVEL < 2
    for (hioffs = 0; hioffs < 256; hioffs++)
#endif
	for (bnr = start; bnr < start+size; bnr++) 
	    membanks[bnr + hioffs * 256] = *bank;
}
